#include "lexer.h"

#include "token.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* helper functions */
char
advance (struct lexer *l)
{
	/* advances the cur pointer */
	/* (consumes the character) */
	if (!l)
		return 0;

	if (l->cur + 1 > buffer_length (l->source))
		return 0;

	char *code = buffer_get (l->source);
	return code[l->cur++];
}

char
peek (struct lexer *l)
{
	/* just takes a peek at the current character */
	/* (doesn't consume the character) */
	if (!l)
		return 0;

	if (l->cur + 1 > buffer_length (l->source))
		return 0;

	char *code = buffer_get (l->source);
	return code[l->cur];
}

char
lookahead (unsigned int n, struct lexer *l)
{
	/* looks at the next character in the source */
	/* does not consume the character */
	if (!l)
		return 0;

	if (l->cur + n > buffer_length (l->source))
		return 0;

	char *code = buffer_get (l->source);
	return code[l->cur + n];
}

_Bool
match (char expected, struct lexer *l)
{
	/* check if cur matches an expectation */
	/* (consumes the character if the match is true) */
	if (!l)
		return 0;

	if (l->cur + 1 > buffer_length (l->source))
		return 0;

	char *code = buffer_get (l->source);
	if (code[l->cur] != expected)
		return 0;

	l->cur++;
	return 1;
}

void
add_token (enum token_type type, struct lexer *l)
{
	if (!l)
		return;

	unsigned int size = l->cur - l->start;
	char *lexeme = calloc (size, sizeof (char));
	strncpy (lexeme, buffer_get (l->source) + l->start, size);

	vector_push_back (token_create_heap (type, l->line, lexeme), &l->tokens);
	free (lexeme);
}

/* public functions */
struct lexer
lexer_create ()
{
	struct lexer l = { 0 };

	l.tokens = vector_create ();
	l.line = 1;

	return l;
}

void
lexer_set_source (struct buffer *b, struct lexer *l)
{
	if (!b || !l)
		return;

	l->source = b;
}

void
lexer_lex (struct lexer *l)
{
	while (l->cur < buffer_length (l->source))
		{
			l->start = l->cur;

			char c = advance (l);

			/* whitespaces */
			if (c == '\n')
				l->line++;
			else if (c == ' ' || c == '\t' || c == '\r')
				continue;
			else if (c == '#')
				while (peek (l) != '\n' && !(l->cur >= buffer_length (l->source)))
					advance (l);
			/* single-character tokens */
			else if (c == '(')
				add_token (TOK_LPAREN, l);
			else if (c == ')')
				add_token (TOK_RPAREN, l);
			else if (c == '{')
				add_token (TOK_LCURLY, l);
			else if (c == '}')
				add_token (TOK_RCURLY, l);
			else if (c == '[')
				add_token (TOK_LSQUAR, l);
			else if (c == ']')
				add_token (TOK_RSQUAR, l);
			else if (c == '.')
				add_token (TOK_DOT, l);
			else if (c == ',')
				add_token (TOK_COMMA, l);
			else if (c == '+')
				add_token (TOK_PLUS, l);
			else if (c == '-')
				add_token (TOK_MINUS, l);
			else if (c == '*')
				add_token (TOK_STAR, l);
			else if (c == '^')
				add_token (TOK_CARET, l);
			else if (c == '/')
				add_token (TOK_SLASH, l);
			else if (c == ';')
				add_token (TOK_SEMICOLON, l);
			else if (c == '?')
				add_token (TOK_QUESTION, l);
			else if (c == '%')
				add_token (TOK_MOD, l);
			/* multi-character tokens */
			else if (c == '=' && match ('=', l))
				add_token (TOK_EQ, l);
			else if (c == '~')
				add_token (match ('=', l) ? TOK_EQ : TOK_NOT, l);
			else if (c == '<')
				add_token (match ('=', l) ? TOK_LE : TOK_LT, l);
			else if (c == '>')
				add_token (match ('=', l) ? TOK_GE : TOK_GT, l);
			else if (c == ':')
				add_token (match ('=', l) ? TOK_ASSIGN : TOK_COLON, l);
			/* check if it's a number, and determine if it's a float or integer */
			/* check if it's a " or ', and get the string */
			/* check if it's an alpha character or _, then handle identifiers */
		}
}

void
lexer_print (struct lexer *l)
{
	puts ("Lexer:");

	for (unsigned int i = 0; i < l->tokens.length; i++)
		{
			struct token *t = l->tokens.elements[i];
			printf ("(TOK_TYPE: %d, \"%s\" - at line %d)\n", t->type,
							buffer_get (&t->lexeme), t->line);
		}
}

void
lexer_free (struct lexer *lexer)
{
	if (!lexer)
		return;

	for (unsigned int i = 0; i < lexer->tokens.length; i++)
		{
			struct token *tok = lexer->tokens.elements[i];
			token_free_heap (tok);
		}

	vector_free (&lexer->tokens);
}
